##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################


from pathlib import Path

from appdirs import AppDirs

_appdirs = AppDirs(appname='F4RATK')


class Cache:
    def __init__(self, directory: Path):
        self._location = directory

    def file(self, name: str) -> Path:
        return self._location.joinpath(name)

    @staticmethod
    def register(location: str) -> 'Cache':
        cache_location = Path(location)
        cache_location.mkdir(parents=True, exist_ok=True)
        return Cache(directory=cache_location)


cache = Cache.register(_appdirs.user_cache_dir)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date
from logging import getLogger

from pandas import DataFrame

from f4ratk.data_reader import fred_reader
from f4ratk.domain import Currency, Frequency
from f4ratk.shared import first_period, last_period

log = getLogger(__name__)


class ExchangeReader:
    EXCHANGE_RATE_COLUMN = 'Exchange Rate'

    def __init__(self, frequency: Frequency):
        self._frequency = frequency

    def exchange_data(
        self, currency: Currency, start: date = None, end: date = None
    ) -> DataFrame:
        symbol = self._symbol(currency)

        data: DataFrame = fred_reader(
            exchange_symbol=symbol, start=start, end=end
        ).read()

        data = data.rename(columns={symbol: self.EXCHANGE_RATE_COLUMN})

        data = self._convert_index_to_periods(data)

        data = data.dropna()

        log.info(f"Exchange data range : {first_period(data)} - {last_period(data)}")

        return data

    def _symbol(self, currency: Currency):
        if currency == Currency.EUR:
            if self._frequency == Frequency.DAILY:
                return 'DEXUSEU'
            elif self._frequency == Frequency.MONTHLY:
                return 'EXUSEU'

        raise NotImplementedError

    def _convert_index_to_periods(self, data):
        frequency = 'B' if self._frequency == Frequency.DAILY else 'M'
        return data.to_period(freq=frequency)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from logging import getLogger
from pathlib import Path
from typing import Iterable, List, Optional

from f4ratk.domain import AnalysisConfig, Currency, Frame, Frequency, Region
from f4ratk.file.reader import FileConfig, ValueFormat
from f4ratk.portfolio.analyze import PortfolioConfiguration
from f4ratk.portfolio.ports import Source
from f4ratk.portfolio.reader.nodes import Analysis, File, Root, Ticker
from f4ratk.ticker.reader import Stock

log = getLogger(__name__)


class TickerSourceReader:
    @staticmethod
    def sources(ticker_nodes: Iterable[dict]) -> List[Source]:
        return [TickerSourceReader._read_ticker_node(node) for node in ticker_nodes]

    @staticmethod
    def _read_ticker_node(ticker_node: dict) -> Source:
        ticker = Ticker(ticker_node)

        log.info(
            "Found ticker source '%s' with %d%% weight.",
            ticker.description,
            ticker.weight,
        )

        source = Source(
            origin=Stock(
                ticker_symbol=ticker.symbol, currency=Currency[ticker.currency]
            ),
            weight=ticker.weight,
        )

        return source


class FileSourceReader:
    @staticmethod
    def sources(file_nodes: Iterable[dict]) -> List[Source]:
        return [FileSourceReader._read_file_node(node) for node in file_nodes]

    @staticmethod
    def _read_file_node(file_node: dict) -> Source:
        file = File(file_node)

        log.info(
            "Found file source '%s' with %d%% weight.",
            file.description,
            file.weight,
        )

        return Source(
            origin=FileConfig(
                path=Path(file.path),
                currency=Currency[file.currency],
                value_format=ValueFormat[file.format],
            ),
            weight=file.weight,
        )


class AnalysisReader:
    @staticmethod
    def analysis(analysis_node: dict) -> AnalysisConfig:
        analysis = Analysis(analysis_node)

        log.info(
            "Found analysis configuration: [%s, %s]",
            analysis.region,
            analysis.frequency,
        )

        config = AnalysisConfig(
            region=Region[analysis.region],
            frame=Frame(frequency=Frequency[analysis.frequency], start=None, end=None),
        )

        return config


def read_requested_document(
    document: dict, name: Optional[str]
) -> Optional[PortfolioConfiguration]:
    root = Root(document)

    if name and name != root.name:
        log.debug("Skipping portfolio configuration '%s'", root.name)
        return

    return _read_document(root)


def _read_document(root: Root) -> Optional[PortfolioConfiguration]:
    log.info("Reading portfolio configuration '%s'", root.name)

    config = AnalysisReader.analysis(root.analysis)

    sources = tuple(
        TickerSourceReader.sources(root.ticker) + FileSourceReader.sources(root.file)
    )

    portfolio = PortfolioConfiguration(sources=sources, config=config)

    _verify_weight(portfolio)

    return portfolio


def _verify_weight(portfolio: PortfolioConfiguration) -> None:
    weight = portfolio.total_weight()
    if weight != 100:
        log.warning("Total portfolio weight of '%d %%' should be 100%%.", weight)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date
from pathlib import Path
from textwrap import dedent
from typing import Optional

from click import Choice, Context, Option, Path as PathArg, argument, group, option

from f4ratk import __VERSION__
from f4ratk.cli.types import Date, RegionChoice
from f4ratk.converter import SourceType, convert_file
from f4ratk.domain import AnalysisConfig, Currency, Frame, Frequency, Region
from f4ratk.file.analyze import analyze_file
from f4ratk.file.reader import FileConfig, ValueFormat
from f4ratk.history import History
from f4ratk.infrastructure import configure_logging, di, instantiate_dependencies
from f4ratk.portfolio.api import PortfolioAnalyzer, PortfolioRequest
from f4ratk.ticker.analyze import analyze_ticker_symbol
from f4ratk.ticker.reader import Stock


def _print_program_info_and_exit(context: Context, _: Option, value: bool) -> None:
    if not value:
        return

    print(
        dedent(
            f"""\
            F4RATK, version {__VERSION__}
            Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>

            https://codeberg.org/toroettg/F4RATK

            F4RATK is free software: you can redistribute it and/or modify
            it under the terms of the GNU Affero General Public License version 3
            as published by the Free Software Foundation.

            This program is distributed in the hope that it will be useful,
            but WITHOUT ANY WARRANTY; without even the implied warranty of
            MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
            GNU Affero General Public License for more details.
            """
        )
    )
    context.exit()


@group()
@option('-v', '--verbose', is_flag=True, help="Increase output verbosity.")
@option(
    '--about',
    callback=_print_program_info_and_exit,
    is_flag=True,
    expose_value=False,
    help="Display program information and exit.",
)
def main(verbose: bool):
    configure_logging(verbose)
    instantiate_dependencies()


@main.command(name="ticker", help="Analyze a ticker symbol.")
@argument('symbol')
@argument('region', type=RegionChoice())
@argument('currency', type=Choice(('USD', 'EUR'), case_sensitive=False))
@option(
    '--start', type=Date(formats=('%Y-%m-%d',)), help="Start of period under review."
)
@option('--end', type=Date(formats=('%Y-%m-%d',)), help="End of period under review.")
@option(
    '--frequency',
    type=Choice(('DAILY', 'MONTHLY'), case_sensitive=False),
    default='MONTHLY',
    show_default=True,
    help="Conduct analysis with given sample frequency.",
)
def ticker(
    symbol: str,
    currency: str,
    region: Region,
    start: Optional[date],
    end: Optional[date],
    frequency: str = 'MONTHLY',
):
    stock = Stock(ticker_symbol=symbol, currency=Currency[currency])

    analysis_config = AnalysisConfig(
        region=region, frame=Frame(frequency=Frequency[frequency], start=start, end=end)
    )

    analyze_ticker_symbol(
        stock=stock,
        analysis_config=analysis_config,
    )


@main.command(
    name="file",
    short_help="Analyze a CSV file.",
    help="""
         Analyze a CSV file.

         Expects a row to be formatted in ISO 8601 with optional day for monthly input
         at the first column; and percentage in US notation with arbitrary-precision
         decimal number value at the second column, where the value may represent
         either prices or returns in percentage:

         'YYYY-MM[-DD],####.####', e.g.,

         '2020-09-11,3.53'
         '2020-09,-0.282'
         """,
)
@argument('path', type=PathArg(dir_okay=False))
@argument('region', type=RegionChoice())
@argument('currency', type=Choice(('USD', 'EUR'), case_sensitive=False))
@argument('value_format', type=Choice(('PRICE', 'RETURN'), case_sensitive=False))
@option(
    '--start', type=Date(formats=('%Y-%m-%d',)), help="Start of period under review."
)
@option('--end', type=Date(formats=('%Y-%m-%d',)), help="End of period under review.")
@option(
    '--frequency',
    type=Choice(('DAILY', 'MONTHLY'), case_sensitive=False),
    default='MONTHLY',
    show_default=True,
    help="Conduct analysis with given sample frequency.",
)
def file(
    path: str,
    currency: str,
    region: Region,
    value_format: ValueFormat,
    start: Optional[date],
    end: Optional[date],
    frequency: str = 'MONTHLY',
):
    target_frequency = Frequency[frequency]

    file_config = FileConfig(
        path=Path(path),
        currency=Currency[currency],
        value_format=ValueFormat[value_format],
    )

    analysis_config = AnalysisConfig(
        region=region, frame=Frame(frequency=target_frequency, start=start, end=end)
    )

    analyze_file(file_config=file_config, analysis_config=analysis_config)


@main.command(name="convert", help="Convert files to the 'file' command format.")
@argument('type', type=Choice(('MSCI',), case_sensitive=False))
@argument('source', type=PathArg(dir_okay=False, exists=True))
def convert(source: str, type: str):
    source_path = Path(source)
    target_path = source_path.with_suffix('.csv')

    convert_file(source=source_path, target=target_path, source_type=SourceType[type])


@main.command(name="history", help="Display historic factor returns.")
@argument('region', type=RegionChoice())
def history(region: Region):
    di[History].display_history(region=region)


@main.command(
    name="portfolio",
    short_help="Analyze a portfolio file.",
    help="""
         Analyze a portfolio file.

         A portfolio consists of a set of multiple sources (e.g.,
         files or ticker symbols) and their respective weights.
         The data for each source is being read, aggregated and
         finally tested against a given analysis configuration.

         Each of those parameters is being read from a file at
         the given path parameter. For details of the required
         format of this file, please see the documentation.
         """,
)
@argument('path', type=PathArg(dir_okay=False, exists=True))
@option(
    '--name',
    type=str,
    help=(
        "Name of the portfolio to analyze in the file at the given path."
        " [default: the first portfolio of the file]"
    ),
)
def portfolio(path: str, name: Optional[str] = None):
    request = PortfolioRequest(path=Path(path), name=name)
    di[PortfolioAnalyzer].analyze_portfolio_file(request=request)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from pandas import DataFrame

from f4ratk.analyze.api import DataAnalyzer
from f4ratk.analyze.evaluation import Critic, EvaluatedResults
from f4ratk.analyze.output import ResultsWriter
from f4ratk.analyze.regression import RegressionRunner, Results
from f4ratk.history import AnnualizedReturns


class DataAnalyzerAdapter(DataAnalyzer):
    def __init__(
        self, regression_runner: RegressionRunner, critic: Critic, writer: ResultsWriter
    ):
        self._regression_runner = regression_runner
        self._critic = critic
        self._writer = writer

    def analyze(
        self,
        returns: DataFrame,
        fama_data: DataFrame,
        historic_returns: AnnualizedReturns,
    ) -> None:
        regression_results = self._regression_runner.run(
            returns=returns, fama_data=fama_data
        )

        evaluated_results = self._evaluate(regression_results, historic_returns)

        self._writer.write(evaluated_results)

    def _evaluate(
        self, regression: Results, historic_returns: AnnualizedReturns
    ) -> EvaluatedResults:
        evaluated_ff6 = self._critic.evaluate(regression.ff6, historic_returns)

        return EvaluatedResults(
            capm=regression.capm,
            ff3=regression.ff3,
            ff5=regression.ff5,
            ff6=evaluated_ff6,
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from f4ratk.analyze.api import DataAnalyzer
from f4ratk.domain import AnalysisConfig
from f4ratk.fama import FamaReader
from f4ratk.history import History
from f4ratk.infrastructure import di
from f4ratk.shared import Normalizer
from f4ratk.ticker.reader import Stock, read_ticker


def analyze_ticker_symbol(stock: Stock, analysis_config: AnalysisConfig) -> None:
    fama_reader = FamaReader(Normalizer())
    analyzer: DataAnalyzer = di[DataAnalyzer]
    history: History = di[History]

    data = read_ticker(stock=stock, frame=analysis_config.frame)

    historic_returns = history.annualized_returns(region=analysis_config.region)

    analyzer.analyze(
        data,
        fama_reader.fama_data(
            region=analysis_config.region, frequency=analysis_config.frame.frequency
        ),
        historic_returns,
    )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from logging import getLogger
from typing import Iterable

from pandas import DataFrame

from f4ratk.analyze.api import DataAnalyzer
from f4ratk.domain import Frame, Frequency, Region
from f4ratk.fama import FamaReader
from f4ratk.file.reader import FileConfig, read_file
from f4ratk.history import History
from f4ratk.portfolio.api import PortfolioAnalyzer
from f4ratk.portfolio.ports import (
    Origin,
    PortfolioConfiguration,
    PortfolioReader,
    PortfolioRequest,
    Source,
)
from f4ratk.shared import format_range
from f4ratk.ticker.reader import Stock, read_ticker

log = getLogger(__name__)


class PortfolioAnalyzerAdapter(PortfolioAnalyzer):
    def __init__(
        self,
        portfolio_reader: PortfolioReader,
        fama_reader: FamaReader,
        analyzer: DataAnalyzer,
        history: History,
    ):
        self._portfolio_reader = portfolio_reader
        self._fama_reader = fama_reader
        self._analyzer = analyzer
        self._history = history

    def analyze_portfolio_file(self, request: PortfolioRequest) -> None:
        portfolio = self._portfolio_reader.read(request=request)

        self._analyze_portfolio(portfolio)

    def _analyze_portfolio(self, portfolio: PortfolioConfiguration) -> None:
        frame = portfolio.config.frame
        region = portfolio.config.region

        portfolio_data = self._read_combined_sources(portfolio.sources, frame)

        fama_data = self._read_fama_data(region, frame.frequency)

        historic_returns = self._history.annualized_returns(region=region)

        self._analyzer.analyze(
            portfolio_data,
            fama_data,
            historic_returns,
        )

    def _read_combined_sources(
        self, sources: Iterable[Source], frame: Frame
    ) -> DataFrame:
        combined = sum(self._read_weighted_data(source, frame) for source in sources)
        return combined.dropna()

    def _read_weighted_data(self, source: Source, frame: Frame) -> DataFrame:
        data = self._read_data(source.origin, frame)
        log.debug("Source data range: %s", format_range(data))
        return data.mul(source.weight / 100)

    @staticmethod
    def _read_data(origin: Origin, frame: Frame) -> DataFrame:
        if isinstance(origin, Stock):
            return read_ticker(stock=origin, frame=frame)
        elif isinstance(origin, FileConfig):
            return read_file(file_config=origin, frame=frame)

        raise ValueError

    def _read_fama_data(self, region: Region, frequency: Frequency) -> DataFrame:
        return self._fama_reader.fama_data(region=region, frequency=frequency)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import logging as log
from dataclasses import dataclass
from enum import Enum, unique

from pandas import NA, DataFrame, DatetimeIndex, merge

from f4ratk.data_reader import fama_french_reader
from f4ratk.domain import Frequency, Region
from f4ratk.shared import Normalizer


@unique
class ReturnsData(Enum):
    DEVELOPED_5_DAILY = 'Developed_5_Factors_Daily'
    DEVELOPED_5_MONTHLY = 'Developed_5_Factors'
    DEVELOPED_MOM_DAILY = 'Developed_Mom_Factor_Daily'
    DEVELOPED_MOM_MONTHLY = 'Developed_Mom_Factor'

    DEVELOPED_EX_US_5_DAILY = 'Developed_ex_US_5_Factors_Daily'
    DEVELOPED_EX_US_5_MONTHLY = 'Developed_ex_US_5_Factors'
    DEVELOPED_EX_US_MOM_DAILY = 'Developed_ex_US_Mom_Factor_Daily'
    DEVELOPED_EX_US_MOM_MONTHLY = 'Developed_ex_US_Mom_Factor'

    US_5_DAILY = 'F-F_Research_Data_5_Factors_2x3_daily'
    US_5_MONTHLY = 'F-F_Research_Data_5_Factors_2x3'
    US_MOM_DAILY = 'F-F_Momentum_Factor_daily'
    US_MOM_MONTHLY = 'F-F_Momentum_Factor'

    EU_5_DAILY = 'Europe_5_Factors_Daily'
    EU_5_MONTHLY = 'Europe_5_Factors'
    EU_MOM_DAILY = 'Europe_Mom_Factor_Daily'
    EU_MOM_MONTHLY = 'Europe_Mom_Factor'

    EMERGING_5_MONTHLY = 'Emerging_5_Factors'
    EMERGING_MOM_MONTHLY = 'Emerging_MOM_Factor'


@dataclass
class ReturnsRequest:
    frequency: Frequency
    five: ReturnsData
    momentum: ReturnsData

    @staticmethod
    def _source_for(region: Region, frequency: Frequency) -> (ReturnsData, ReturnsData):
        if region == Region.DEVELOPED:
            if frequency == Frequency.DAILY:
                return (
                    ReturnsData.DEVELOPED_5_DAILY,
                    ReturnsData.DEVELOPED_MOM_DAILY,
                )
            elif frequency == Frequency.MONTHLY:
                return (
                    ReturnsData.DEVELOPED_5_MONTHLY,
                    ReturnsData.DEVELOPED_MOM_MONTHLY,
                )
        if region == Region.DEVELOPED_EX_US:
            if frequency == Frequency.DAILY:
                return (
                    ReturnsData.DEVELOPED_EX_US_5_DAILY,
                    ReturnsData.DEVELOPED_EX_US_MOM_DAILY,
                )
            elif frequency == Frequency.MONTHLY:
                return (
                    ReturnsData.DEVELOPED_EX_US_5_MONTHLY,
                    ReturnsData.DEVELOPED_EX_US_MOM_MONTHLY,
                )
        elif region == Region.US:
            if frequency == Frequency.DAILY:
                return (ReturnsData.US_5_DAILY, ReturnsData.US_MOM_DAILY)
            elif frequency == Frequency.MONTHLY:
                return (ReturnsData.US_5_MONTHLY, ReturnsData.US_MOM_MONTHLY)
        elif region == Region.EU:
            if frequency == Frequency.DAILY:
                return (ReturnsData.EU_5_DAILY, ReturnsData.EU_MOM_DAILY)
            elif frequency == Frequency.MONTHLY:
                return (ReturnsData.EU_5_MONTHLY, ReturnsData.EU_MOM_MONTHLY)
        elif region == Region.EMERGING:
            if frequency == Frequency.MONTHLY:
                return (
                    ReturnsData.EMERGING_5_MONTHLY,
                    ReturnsData.EMERGING_MOM_MONTHLY,
                )

        raise ValueError(
            f"'{frequency.name}' data unvailable for '{region.name}' region."
        )

    @staticmethod
    def of(region: Region, frequency: Frequency) -> 'ReturnsRequest':
        five, momentum = ReturnsRequest._source_for(region, frequency)

        return ReturnsRequest(frequency=frequency, five=five, momentum=momentum)


class FamaReader:
    def __init__(self, normalizer: Normalizer):
        self._normalizer = normalizer

    def fama_data(self, region: Region, frequency: Frequency) -> DataFrame:
        sources = ReturnsRequest.of(region, frequency)
        return self._fama_data(sources)

    def _fama_data(
        self,
        request: ReturnsRequest,
    ) -> DataFrame:
        data_ff = self._fama_ff_data(request.five, request.frequency)
        data_mom = self._fama_momentum_data(request.momentum, request.frequency)

        combined: DataFrame = merge(
            data_ff, data_mom, left_index=True, right_index=True
        )

        log.debug(f"Fama data of set '{combined}' ends at\n%s", combined.tail())

        return combined

    def _fama_ff_data(self, source: ReturnsData, frequency: Frequency) -> DataFrame:
        data = self._load_fama_data(source, frequency)
        data = data.rename(columns={'Mkt-RF': 'MKT'})
        return data

    def _fama_momentum_data(
        self, source: ReturnsData, frequency: Frequency
    ) -> DataFrame:
        data: DataFrame = self._load_fama_data(source, frequency)

        if source == ReturnsData.US_MOM_DAILY or source == ReturnsData.US_MOM_MONTHLY:
            data = data.rename(columns={'Mom   ': 'WML'})

        return data

    def _load_fama_data(self, source: ReturnsData, frequency: Frequency) -> DataFrame:
        data: DataFrame = fama_french_reader(returns_data=source.value).read()[0]

        if isinstance(data.index, DatetimeIndex):
            data = self._normalizer.index_to_periods(data, frequency)
            log.debug(
                f"Fama reader returned DatetimeIndex for source '{source}', converted to frequency '{frequency}'"  # noqa: E501
            )

        data = data.replace(-99.99, NA)
        data = data.dropna()

        return data



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from datetime import date
from enum import Enum, unique
from pathlib import Path
from typing import Optional

from pandas import DataFrame, merge

from f4ratk.data_reader import CsvFileReader
from f4ratk.domain import Currency, Frame, Frequency
from f4ratk.exchange import ExchangeReader
from f4ratk.shared import Normalizer


@unique
class ValueFormat(Enum):
    PRICE = 'Prices'
    RETURN = 'Returns'


@dataclass(frozen=True)
class FileConfig:
    path: Path
    currency: Currency
    value_format: ValueFormat


def read_file(file_config: FileConfig, frame: Frame) -> DataFrame:
    file_reader = FileReader(
        csv_reader=CsvFileReader(path=file_config.path),
        exchange_reader=ExchangeReader(frame.frequency),
        currency=file_config.currency,
        value_format=file_config.value_format,
        normalizer=Normalizer(),
    )

    return file_reader.read(start=frame.start, end=frame.end, frequency=frame.frequency)


class FileReader:
    def __init__(
        self,
        csv_reader: CsvFileReader,
        exchange_reader: ExchangeReader,
        currency: Currency,
        value_format: ValueFormat,
        normalizer: Normalizer,
    ):
        self._csv_reader = csv_reader
        self._exchange_reader = exchange_reader
        self._currency = currency
        self._value_format = value_format
        self._normalizer = normalizer

    def read(
        self, start: Optional[date], end: Optional[date], frequency: Frequency
    ) -> DataFrame:
        data = self._csv_reader.read().sort_index()

        data = self._normalizer.index_to_periods(data=data, frequency=frequency)

        if self._currency is not Currency.USD:
            data = self._convert_currency(
                data=data, currency=self._currency, start=start, end=end
            )

        if self._value_format == ValueFormat.PRICE:
            data = data[['Returns']].pct_change() * 100

        return data.dropna()

    def _convert_currency(
        self, data: DataFrame, currency: Currency, start: date, end: date
    ) -> DataFrame:
        exchange_data = self._exchange_reader.exchange_data(currency, start, end)

        data = merge(data, exchange_data, left_index=True, right_index=True)

        data['Returns'] = data['Returns'] * data[ExchangeReader.EXCHANGE_RATE_COLUMN]

        return data[['Returns']]



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from logging import getLogger
from pathlib import Path
from typing import Any, Dict, Generator, Hashable, List, Optional, Union

from yaml import safe_load_all

from f4ratk.portfolio.ports import (
    PortfolioConfiguration,
    PortfolioReader,
    PortfolioRequest,
)
from f4ratk.portfolio.reader.reader import read_requested_document

log = getLogger(__name__)


class PortfolioReaderAdapter(PortfolioReader):
    def read(self, request: PortfolioRequest) -> PortfolioConfiguration:
        documents = self._load_file_content(request.path)

        configurations = (
            read_requested_document(document, request.name) for document in documents
        )

        configuration = self._find_first(configurations)

        return configuration

    @staticmethod
    def _load_file_content(
        path: Path,
    ) -> List[Union[Dict[Hashable, Any], list]]:
        with open(path, 'r') as handle:
            content = list(safe_load_all(handle))
            content = list(content) if isinstance(content, dict) else content
        return content

    @staticmethod
    def _find_first(
        configurations: Generator[Optional[PortfolioConfiguration], None, None],
    ) -> PortfolioConfiguration:
        try:
            return next(config for config in configurations if config is not None)
        except StopIteration:
            raise SystemExit("Unable to read requested portfolio configuration.")



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from logging import getLogger
from math import sqrt

from pandas import DataFrame, Period, Series, concat

from f4ratk.domain import Frequency, Region
from f4ratk.fama import FamaReader
from f4ratk.shared import first_period, format_range, last_period

log = getLogger(__name__)


@dataclass(frozen=True)
class AnnualizedReturns:
    _data: DataFrame
    _first: Period
    _last: Period

    def all(self) -> Series:
        return self._data['Returns']

    def __str__(self) -> str:
        sep = f"\n{'#' * 25}\n"
        nl = '\n'

        header = "Annualized Returns (%)"
        period = f"Period: {self._first} - {self._last}"
        frequency = f"Frequency: {Frequency.MONTHLY.name}"

        content = self._data.to_string(col_space=10, float_format='{:0.3f}'.format)

        return "".join((header, sep, period, nl, frequency, nl * 2, content))


class Historian:
    def __init__(self, fama_reader: FamaReader):
        self._fama_reader = fama_reader

    def annualized_returns(self, region: Region) -> AnnualizedReturns:
        returns = self._historic_returns(region=region)
        annualized = self._monthly_annualized_returns(data=returns)

        log.info(f"Historic data range {format_range(returns)}")

        return annualized

    def _historic_returns(self, region: Region) -> DataFrame:
        return self._fama_reader.fama_data(region=region, frequency=Frequency.MONTHLY)

    @staticmethod
    def _monthly_annualized_returns(data: DataFrame) -> AnnualizedReturns:
        periods = 12 / len(data)

        annualized_returns: Series = Series(
            data.div(100).add(1).prod().pow(periods).sub(1).mul(100), name="Returns"
        )

        annualized_std: Series = Series(data.std().mul(sqrt(12)), name="SD")

        return AnnualizedReturns(
            _data=concat((annualized_returns, annualized_std), axis='columns'),
            _first=first_period(data),
            _last=last_period(data),
        )


class History:
    def __init__(self, historian: Historian):
        self._historian = historian

    def display_history(self, region: Region):
        returns = self.annualized_returns(region)
        print(returns)

    def annualized_returns(self, region: Region):
        return self._historian.annualized_returns(region=region)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from datetime import date
from logging import getLogger
from typing import Optional

from pandas import DataFrame, merge

from f4ratk.data_reader import yahoo_reader
from f4ratk.domain import Currency, Frame, Frequency
from f4ratk.exchange import ExchangeReader
from f4ratk.shared import Downsampler, Normalizer, QualityChecker

log = getLogger(__name__)


@dataclass(frozen=True)
class Stock:
    ticker_symbol: str
    currency: Currency = Currency.USD


def read_ticker(stock: Stock, frame: Frame) -> DataFrame:
    ticker_reader = TickerReader(
        ExchangeReader(frame.frequency), Normalizer(), Downsampler(), QualityChecker()
    )

    return ticker_reader.data(stock, frame.frequency, frame.start, frame.end)


class TickerReader:
    ADJUSTED_CLOSE_COLUMN = 'Adj Close'

    def __init__(
        self,
        exchange_reader: ExchangeReader,
        normalizer: Normalizer,
        downsampler: Downsampler,
        checker: Optional[QualityChecker] = None,
    ):
        self._exchange_reader = exchange_reader
        self._normalizer = normalizer
        self._downsampler = downsampler
        self._checker = checker

    def data(
        self,
        stock: Stock,
        frequency: Frequency,
        start: Optional[date] = None,
        end: Optional[date] = None,
    ) -> DataFrame:
        data = self._read_stock_data(stock, start, end)
        data = self._normalize_date_range(data, frequency)

        if stock.currency is not Currency.USD:
            data = self._convert_currency(data, stock.currency, start, end)

        data['Returns'] = self._calculate_returns_as_relative_percentage(data)

        log.debug(
            f"Stock data of symbol '{stock.ticker_symbol}' starts at: \n%s", data.head()
        )

        return data[['Returns']].dropna()

    def _read_stock_data(
        self, stock: Stock, start: Optional[date], end: Optional[date]
    ) -> DataFrame:
        data: DataFrame = yahoo_reader(stock.ticker_symbol, start, end).read()
        return data

    def _convert_currency(
        self,
        data: DataFrame,
        currency: Currency,
        start: Optional[date],
        end: Optional[date],
    ) -> DataFrame:
        exchange_data = self._exchange_reader.exchange_data(currency, start, end)

        data = merge(data, exchange_data, left_index=True, right_index=True)

        source_currency_close_column = f"{self.ADJUSTED_CLOSE_COLUMN} ({currency.name})"

        data = data.rename(
            columns={self.ADJUSTED_CLOSE_COLUMN: source_currency_close_column}
        )

        data[self.ADJUSTED_CLOSE_COLUMN] = (
            data[source_currency_close_column]
            * data[ExchangeReader.EXCHANGE_RATE_COLUMN]
        )

        return data

    def _calculate_returns_as_relative_percentage(self, data: DataFrame) -> DataFrame:
        return data[[self.ADJUSTED_CLOSE_COLUMN]].pct_change() * 100

    def _normalize_date_range(self, data: DataFrame, frequency: Frequency) -> DataFrame:
        data = self._normalizer.index_to_periods(data=data, frequency=Frequency.DAILY)

        if self._checker:
            self._checker.investigate(data)

        if frequency == Frequency.MONTHLY:
            data = self._down_sample(data)

        return data

    def _down_sample(self, data: DataFrame) -> DataFrame:
        log.info("Downsampling ticker data to monthly rate")
        data = self._downsampler.monthly_sample(data=data)
        return data



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from logging import getLogger
from typing import Union

from pandas import DataFrame, Period, period_range
from pandas.tseries.offsets import BusinessDay, MonthEnd

from f4ratk.domain import Frequency

log = getLogger(__name__)


def format_range(df: DataFrame) -> str:
    return f"{first_period(df)} - {last_period(df)}"


def last_period(df: DataFrame) -> Period:
    return df.index.max()


def first_period(df: DataFrame) -> Period:
    return df.index.min()


class Downsampler:
    def monthly_sample(self, data: DataFrame) -> DataFrame:
        return data.resample(BusinessDay()).interpolate().resample(MonthEnd()).last()


class Normalizer:
    def index_to_periods(self, data: DataFrame, frequency: Frequency) -> DataFrame:
        offset = self._to_offset(frequency)
        return data.to_period(freq=offset)

    def _to_offset(self, frequency: Frequency) -> Union[BusinessDay, MonthEnd]:
        if frequency == Frequency.DAILY:
            return BusinessDay()
        elif frequency == Frequency.MONTHLY:
            return MonthEnd()

        raise NotImplementedError


class QualityChecker:
    def investigate(self, data: DataFrame):
        observation_pct = self._relative_observations(data)

        if observation_pct < 0.95:
            log.warning(
                f"Quote data contains only {observation_pct:.2%} of possible periods"
            )
        else:
            log.info(f"Quote data contains {observation_pct:.2%} of possible periods")

    @staticmethod
    def _relative_observations(data: DataFrame) -> float:
        return len(data) / len(
            period_range(start=first_period(data), end=last_period(data))
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from logging import getLogger
from typing import Iterator

from pandas import Series
from statsmodels.regression.linear_model import RegressionResultsWrapper

from f4ratk.analyze.regression import ModelType, Result
from f4ratk.history import AnnualizedReturns

log = getLogger(__name__)


@dataclass(frozen=True)
class EvaluatedResult:
    model_type: ModelType
    model: RegressionResultsWrapper
    evaluation: float

    @staticmethod
    def of(result: Result, evaluation: float) -> "EvaluatedResult":
        return EvaluatedResult(
            model_type=result.model_type, model=result.model, evaluation=evaluation
        )


@dataclass(frozen=True)
class EvaluatedResults:
    capm: Result
    ff3: Result
    ff5: Result
    ff6: EvaluatedResult

    def __iter__(self) -> Iterator[Result]:
        yield from (self.capm, self.ff3, self.ff5, self.ff6)


class Critic:
    _TARGET_FACTORS = ['MKT', 'SMB', 'HML', 'RMW', 'CMA', 'WML']

    def evaluate(
        self, result: Result, historic_returns: AnnualizedReturns
    ) -> EvaluatedResult:
        log.debug(f"Evaluating with: {historic_returns}")

        factors = self._normalize(self._significant_factors(result.model))
        expectation = self._expectation(historic_returns)

        forecast = factors.mul(expectation).sum()

        return EvaluatedResult.of(result=result, evaluation=forecast)

    def _significant_factors(self, model: RegressionResultsWrapper) -> Series:
        targeted = model.params[Critic._TARGET_FACTORS]
        return targeted[model.pvalues < 0.1]

    def _normalize(self, factors: Series) -> Series:
        if 'MKT' in factors:
            factors['MKT'] -= 1
        return factors

    def _expectation(self, historic_returns: AnnualizedReturns) -> Series:
        return historic_returns.all().div(2)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from textwrap import dedent
from typing import TextIO

from statsmodels.iolib.summary2 import summary_col

from f4ratk.analyze.evaluation import EvaluatedResult, EvaluatedResults


class ResultsWriter:
    def __init__(self, file: TextIO):
        self._file = file

    def write(self, results: EvaluatedResults) -> None:
        self._write_summary(results)
        print(file=self._file)
        self._write_forecast(results.ff6)
        print(file=self._file)
        self._write_ff6_details(results.ff6)

    def _write_summary(self, results: EvaluatedResults) -> None:
        summary = summary_col(
            results=list(result.model for result in results),
            stars=True,
            model_names=list(result.model_type.value for result in results),
            info_dict={
                'N': lambda x: "{0:d}".format(int(x.nobs)),
            },
            regressor_order=['Intercept', 'MKT', 'SMB', 'HML', 'RMW', 'CMA', 'WML'],
        )

        print(summary, file=self._file)

    def _write_forecast(self, result: EvaluatedResult) -> None:
        print(
            dedent(
                f"""\
                Expected excess return before cost: {result.evaluation:.3}%

                Notes:
                  Based on the determined factor loadings and historic
                  factor returns. Past performance is not a guarantee of
                  future results nor an indicator of future performance.

                """
            ),
            file=self._file,
        )

    def _write_ff6_details(self, ff6: EvaluatedResult) -> None:
        print(ff6.model.summary(), file=self._file)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from enum import Enum, unique
from logging import getLogger

import pandas
from pandas import DataFrame
from statsmodels.formula import api as sm
from statsmodels.regression.linear_model import RegressionResultsWrapper

from f4ratk.shared import first_period, last_period

log = getLogger(__name__)


@unique
class ModelType(Enum):
    CAPM = "CAPM"
    FF3 = "FF3"
    FF5 = "FF5"
    FF6 = "FF6"


@dataclass(frozen=True)
class Result:
    model_type: ModelType
    model: RegressionResultsWrapper


@dataclass(frozen=True)
class Results:
    capm: Result
    ff3: Result
    ff5: Result
    ff6: Result


class RegressionRunner:
    def run(self, returns: DataFrame, fama_data: DataFrame) -> Results:
        combined = self._combine(returns, fama_data)

        return Results(
            capm=self._run(model_type=ModelType.CAPM, data=combined),
            ff3=self._run(model_type=ModelType.FF3, data=combined),
            ff5=self._run(model_type=ModelType.FF5, data=combined),
            ff6=self._run(model_type=ModelType.FF6, data=combined),
        )

    def _combine(self, returns: DataFrame, fama_data: DataFrame) -> DataFrame:
        log.info(
            f"Returns data range: {first_period(returns)} - {last_period(returns)}"
        )
        log.info(
            f"Fama data range : {first_period(fama_data)} - {last_period(fama_data)}"
        )

        combined: DataFrame = pandas.merge(
            returns, fama_data, left_index=True, right_index=True
        )

        combined['XsRet'] = combined['Returns'] - combined['RF']

        log.info(
            f"Result date range: {first_period(combined)} - {last_period(combined)}"
        )

        return combined

    def _run(self, model_type: ModelType, data: DataFrame) -> Result:
        formula = self._formula(model_type)
        model = self._model(formula=formula, data=data)
        return Result(model_type=model_type, model=model)

    def _formula(self, type: ModelType) -> str:
        if type == ModelType.CAPM:
            return 'XsRet ~ MKT'
        elif type == ModelType.FF3:
            return 'XsRet ~ MKT + SMB + HML'
        elif type == ModelType.FF5:
            return 'XsRet ~ MKT + SMB + HML + RMW + CMA'
        elif type == ModelType.FF6:
            return 'XsRet ~ MKT + SMB + HML + RMW + CMA + WML'
        raise ValueError

    def _model(self, formula: str, data: DataFrame) -> RegressionResultsWrapper:
        return sm.ols(formula=formula, data=data).fit(
            cov_type='HAC', cov_kwds={'maxlags': None}, use_t=True
        )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from typing import List


class Node:
    def __init__(self, node: dict):
        self._node = node

    @staticmethod
    def _normalize(value: str) -> str:
        uppercase = value.upper()
        return uppercase.replace('-', '_')

    def _read(self, key: str) -> str:
        return self._node[key]

    def _read_normalized(self, key: str) -> str:
        raw = self._read(key)
        return self._normalize(raw)


class Root(Node):
    @property
    def name(self) -> str:
        return self._read('name')

    @property
    def file(self) -> List[dict]:
        return self._node.get('files', [])

    @property
    def ticker(self) -> List[dict]:
        return self._node.get('tickers', [])

    @property
    def analysis(self) -> dict:
        return self._node['analysis']


class Source(Node):
    @property
    def description(self) -> str:
        return self._read('description')

    @property
    def currency(self) -> str:
        return self._read_normalized('currency')

    @property
    def weight(self) -> int:
        return int(self._node['weight'])


class Ticker(Source):
    @property
    def symbol(self) -> str:
        return self._read_normalized('symbol')


class File(Source):
    @property
    def path(self) -> str:
        return self._read('path')

    @property
    def format(self) -> str:
        return self._read_normalized('format')


class Analysis(Node):
    @property
    def region(self) -> str:
        return self._read_normalized('region')

    @property
    def frequency(self) -> str:
        return self._read_normalized('frequency')



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date, datetime
from typing import Optional

from click import Choice, DateTime

from f4ratk.domain import Region


class Date(DateTime):
    name = "date"

    def __init__(self, formats=None):
        formats = formats or ["%Y-%m-%d"]
        super().__init__(formats)

    def convert(self, value, param, ctx) -> Optional[date]:
        converted: Optional[datetime] = super().convert(value, param, ctx)
        return converted.date() if converted else None

    def __repr__(self):
        return 'Date'


class RegionChoice(Choice):
    name = "regionchoice"

    def __init__(self):
        super().__init__(
            choices=('DEVELOPED', 'DEVELOPED-EX-US', 'US', 'EU', 'EMERGING'),
            case_sensitive=False,
        )

    def convert(self, value, param, ctx) -> Optional[Region]:
        converted: Optional[str] = super().convert(value, param, ctx)
        return Region[converted.replace('-', '_')] if converted else None



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

import logging as log
from sys import stdout
from typing import Any, Dict, Type

from f4ratk.analyze.analyze import DataAnalyzerAdapter
from f4ratk.analyze.api import DataAnalyzer
from f4ratk.analyze.evaluation import Critic
from f4ratk.analyze.output import ResultsWriter
from f4ratk.analyze.regression import RegressionRunner
from f4ratk.fama import FamaReader
from f4ratk.history import Historian, History
from f4ratk.portfolio.analyze import PortfolioAnalyzerAdapter
from f4ratk.portfolio.api import PortfolioAnalyzer
from f4ratk.portfolio.reader.api import PortfolioReaderAdapter
from f4ratk.shared import Normalizer


def configure_logging(verbose: bool) -> None:
    log.basicConfig(level=log.DEBUG if verbose else log.INFO)
    log.getLogger("urllib3").setLevel(log.INFO)


di: Dict[Type[Any], Any] = dict()


def instantiate_dependencies() -> None:
    log.debug("Bootstrapping application dependencies")
    global di

    fama_reader: FamaReader = FamaReader(Normalizer())

    data_analyer: DataAnalyzer = DataAnalyzerAdapter(
        regression_runner=RegressionRunner(),
        critic=Critic(),
        writer=ResultsWriter(file=stdout),
    )

    history: History = History(historian=Historian(fama_reader=fama_reader))

    di.update(
        {
            DataAnalyzer: data_analyer,
            History: history,
            PortfolioAnalyzer: PortfolioAnalyzerAdapter(
                portfolio_reader=PortfolioReaderAdapter(),
                fama_reader=fama_reader,
                analyzer=data_analyer,
                history=history,
            ),
        }
    )



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from enum import Enum, unique
from logging import getLogger
from pathlib import Path

from pandas import DataFrame, ExcelFile

from f4ratk.domain import Frequency
from f4ratk.shared import Normalizer, first_period, last_period

log = getLogger(__name__)


@unique
class SourceType(Enum):
    MSCI = "MSCI Index History"


class MSCIConverter:
    _HEADER = ('Dates', 'Returns')

    def __init__(self, normalizer: Normalizer):
        self._normalizer = normalizer

    def convert(self, source: Path, target: Path):
        data = self._read_msci_file(source)

        log.info(
            f"Content date ranges from '{first_period(data)}' to '{last_period(data)}'"
        )

        self._write_file(target, data)

    def _read_msci_file(self, path: Path) -> DataFrame:
        log.info(f"Reading input file'{path}'")
        document = ExcelFile(path)
        sheet = document.parse(
            parse_dates=True, names=self._HEADER, skiprows=6, skipfooter=18, index_col=0
        )

        data = self._normalizer.index_to_periods(sheet, Frequency.MONTHLY)
        return data

    @staticmethod
    def _write_file(target: Path, data: DataFrame) -> None:
        log.info(f"Writing output file'{target}'")
        data.to_csv(target, header=False)


def convert_file(source: Path, target: Path, source_type: SourceType) -> None:
    MSCIConverter(normalizer=Normalizer()).convert(source, target)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from datetime import date, datetime, timedelta
from pathlib import Path
from typing import Optional

import requests_cache
from pandas import DataFrame, read_csv
from pandas_datareader.famafrench import FamaFrenchReader
from pandas_datareader.fred import FredReader
from pandas_datareader.yahoo.daily import YahooDailyReader

from f4ratk.directories import cache


def _cached_session() -> requests_cache.CachedSession:
    cache_duration = timedelta(days=14)
    cache_location = str(cache.file(name='requests'))

    # noinspection PyTypeChecker
    session = requests_cache.CachedSession(
        cache_name=cache_location, backend='sqlite', expire_after=cache_duration
    )
    session.cache.remove_old_entries(datetime.utcnow() - cache_duration)

    return session


_session = _cached_session()


def fama_french_reader(returns_data: str) -> FamaFrenchReader:
    return FamaFrenchReader(symbols=returns_data, session=_session, start='1920')


def yahoo_reader(
    ticker_symbol: str, start: Optional[date], end: Optional[date]
) -> YahooDailyReader:
    return YahooDailyReader(
        symbols=ticker_symbol,
        start=start if start else '1970',
        end=end,
        session=_session,
    )


def fred_reader(
    exchange_symbol: str, start: Optional[date], end: Optional[date]
) -> FredReader:
    return FredReader(
        symbols=exchange_symbol,
        start=start if start else '1970',
        end=end,
        session=_session,
    )


class CsvFileReader:
    _HEADER = ('Dates', 'Returns')

    def __init__(self, path: Path):
        self._path = path

    def read(self) -> DataFrame:
        return read_csv(self._path, parse_dates=True, index_col=0, names=self._HEADER)


def csv_reader(path: Path) -> CsvFileReader:
    return CsvFileReader(path=path)



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple, Union

from f4ratk.domain import AnalysisConfig
from f4ratk.file.reader import FileConfig
from f4ratk.ticker.reader import Stock

Origin = Union[Stock, FileConfig]


@dataclass(frozen=True)
class PortfolioRequest:
    path: Path
    name: Optional[str]


@dataclass(frozen=True)
class Source:
    origin: Origin
    weight: int


@dataclass(frozen=True)
class PortfolioConfiguration:
    sources: Tuple[Source, ...]
    config: AnalysisConfig

    def total_weight(self) -> int:
        return sum(source.weight for source in self.sources)


class PortfolioReader(metaclass=ABCMeta):
    @abstractmethod
    def read(self, request: PortfolioRequest) -> PortfolioConfiguration:
        """Read the portfolio configuration matching the given request."""



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from dataclasses import dataclass
from datetime import date
from enum import Enum, unique
from typing import Optional


@unique
class Region(Enum):
    DEVELOPED = 'Developed'
    DEVELOPED_EX_US = 'Developed excluding United States'
    EMERGING = 'Emerging'
    US = 'United States'
    EU = 'Europe'


@unique
class Frequency(Enum):
    DAILY = 'Daily'
    MONTHLY = 'Monthly'


@unique
class Currency(Enum):
    EUR = 'Euro'
    USD = 'US Dollar'


@dataclass(frozen=True)
class Frame:
    frequency: Frequency
    start: Optional[date]
    end: Optional[date]


@dataclass(frozen=True)
class AnalysisConfig:
    region: Region
    frame: Frame



##############################################################################
# Copyright (C) 2020 - 2021 Tobias Röttger <dev@roettger-it.de>
#
# This file is part of F4RATK.
#
# F4RATK is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

from abc import ABCMeta, abstractmethod

from pandas import DataFrame

from f4ratk.history import AnnualizedReturns


class DataAnalyzer(metaclass=ABCMeta):
    @abstractmethod
    def analyze(
        self,
        returns: DataFrame,
        fama_data: DataFrame,
        historic_returns: AnnualizedReturns,
    ) -> None:
        """Analyze data in regards to fit for Fama / French models."""
